/*
  defs: {
    url: String, // required
    itemId: Function (item), // required
    mapId: Function (map), // required
    url: String, // required
    value: Function (item), // required
    ranges: [
      {
        label: String,
        values: [0, 1, 2],
        ticks: ['a', 'b', 'c'],
        colors: ['#000', '#333', '#666']
      }
    ],// required

    showOnUndefined: Function (id),
    groupIds: Array,
    label: Function (item),
    selectedLabel: Function (item),
    striped: Function (item),
    bubble: Function (item),
    bg: String, // url for bg
    download: Boolean
    downloadTitle: String,
    zoomable: Boolean,
    hoverable: Boolean,
    selectable: Boolean,
    unselectable: Boolean,
    legend: Boolean,
    color: Function
    tooltip: Function
  }
*/

const appendOnce = (parent, element, cssClass) => {
  const cssSelector = element + '.' + cssClass
  const child = parent.selectAll(cssSelector).data([0])
  child.enter().append(element).attr('class', cssClass)
  return parent.select(cssSelector)
}

const getOnePixel = (node, width) => {
  const newWidth = node.node().getBoundingClientRect().width
  return width / parseInt(newWidth)
}

const circle = (radius, center) => {
  return `M ${center.x - radius}, ${center.y} a ${radius},${radius} 0 1,0 ${radius * 2},0 a ${radius},${radius} 0 1,0 -${radius * 2},0`
}

const holdInBounds = (bounds, path, center, d) => {
  const rect = path.bounds(bounds)
  const halfWidth = d.bbox.width / 2
  const halfHeight = d.bbox.height / 2
  center.x = Math.max(rect[0][0], center.x - halfWidth) + halfWidth
  center.x = Math.min(rect[1][0], center.x + halfWidth) - halfWidth
  center.y = Math.max(rect[0][1], center.y - halfHeight) + halfHeight
  center.y = Math.min(rect[1][1], center.y + halfHeight) - halfHeight
  return center
}

const center = (bounds, path, defs, d) => {
  const found = bounds.features.find(o => defs.mapId(o.properties) === defs.itemId(d))
  if (!found) {
    return { x: -Infinity, y: -Infinity }
  }
  const center = path.centroid(found.geometry)
  if (!center) {
    return { x: -Infinity, y: -Infinity }
  }
  return { x: center[0], y: center[1] }
}

const bubbleGeometry = (defs, center, d) => {
  return defs.bubbleGeometry ? defs.bubbleGeometry(defs.bubble(d), center) : circle(defs.bubble(d), center)
}

const updateSizes = (d3, node, staticNode, defsNode, data, defs, width, height, id, selectedId) => {
  const onePixel = getOnePixel(node, width)

  const pattern = d3.select('#pattern-stripe')
  if (pattern) {
    pattern
      .attr('height', onePixel * 16)
      .attr('width', onePixel * 16)
    const rects = pattern.selectAll('rect')
    rects
      .attr('height', onePixel * 16)
      .attr('width', onePixel * 8)
      .attr('x', (_, i) => i === 1 ? onePixel * 8 : 0)
  }

  let top = 0
  const download = node.select('.download')
  if (download) {
    download.attr('transform', `translate(${width - 35 * onePixel} ${5 * onePixel}) scale(${onePixel * 1.5})`)
    top = 45
  }

  const plus = node.select('.plus')
  if (plus) {
    plus.attr('transform', `translate(${width - 35 * onePixel} ${(5 + top) * onePixel}) scale(${onePixel * 1.5})`)
  }
  const minus = node.select('.minus')
  if (minus) {
    minus.attr('transform', `translate(${width - 35 * onePixel} ${(37 + top) * onePixel}) scale(${onePixel * 1.5})`)
  }

  if (defs.legend) {
    addLegend(d3, staticNode, defsNode, data, defs, width, height, id, onePixel, selectedId)
  }
}
const screenSize = (d3, node, staticNode, defsNode, data, defs, width, height, id, selectedId) => {
  d3.select(window).on(`resize.id${id}`, function() {
    updateSizes(d3, node, staticNode, defsNode, data, defs, width, height, id, selectedId)
  })
  updateSizes(d3, node, staticNode, defsNode, data, defs, width, height, id, selectedId)
}

const addGradient = (d3, defs, colors, nr, id) => {
  const colorScale = d3
    .scaleQuantize()
    .range(colors)
  appendOnce(defs, 'linearGradient', `gradient-${nr}-${id}`)
    .attr('id', `gradient-${nr}-${id}`)
    .selectAll('stop')
    .data(colorScale.range())
    .enter()
    .append('stop')
    .attr('offset', (d, i) => { return i / (colorScale.range().length - 1) })
    .attr('stop-color', d => { return d })
}

const darkestHexColor = (colors) => {
  return '#' + colors.map(o => {
    const color = o
      .replace('#', '') // remove hash
      .replace(/^(.)(.)(.)$/, '$1$1$2$2$3$3') // 3 to 6
      .replace(/^(.{6})$/, '$1ff') // 6 to 8
    return parseInt(color, 16)
  }).sort((a, b) => a - b)[0]
    .toString(16) // back to hex again
    .padStart(8, '0') // fill up with zeros
    .replace(/ff$/, '') // remove ff on end
}

const addLegend = (d3, node, defsNode, data, defs, width, height, id, onePixel, selectedId) => {
  const fontSize = onePixel * 16

  const margin = 2 * onePixel
  const padding = 5 * onePixel
  const textHeight = fontSize
  const barHeight = 12 * onePixel

  const outerWidth = width - margin * 2
  const outerHeight = padding + textHeight + padding + barHeight + padding + textHeight + padding
  const innerWidth = outerWidth - padding * 2

  const top = height - outerHeight - margin
  const textBottom = textHeight + padding
  const barTop = textBottom + padding
  const left = margin
  const tickTop = barTop + barHeight + padding

  const legend = appendOnce(node, 'g', 'legend')
    .attr('transform', `translate(${left},${top})`)

  // bubble min max size
  if (defs.bubble !== undefined) {
    const biggest = defs.bubble ? data.sort((a, b) => defs.bubble(b) - defs.bubble(a))[0] : null
    let form = null
    let maxVal = 0
    if(biggest) {
      form = bubbleGeometry(defs, { x: 0, y: 0 }, biggest)
      maxVal = defs.bubble(biggest)
    }

    const min = 20
    const max = 40
    const bubbleSteps = defs.bubble ? [min, (min + max) / 2, max] : []

    const bubbleScala = d3
      .scaleLinear()
      .domain(defs.bubble ? d3.extent(data.map(defs.bubble)) : [])
      .range([min, max])

    appendOnce(legend, 'g', 'legendBgBubble')
      .selectAll('path')
      .data(defs.bubble ? [45] : [])
      .join(
        enter => enter
          .append('path')
          .attr('fill', '#ffffffdd')
          .attr('transform', d => `translate(${max}, -${max}) scale(${maxVal ? (1 * d / maxVal) : 1})`)
          .attr('d', form),
        update => update
          .attr('fill', '#ffffffdd')
          .attr('transform', d => `translate(${max}, -${max}) scale(${maxVal ? (1 * d / maxVal) : 1})`)
          .attr('d', form),
        exit => exit
          .remove()
      )
      
    appendOnce(legend, 'g', 'legendBubble')
      .selectAll('path')
      .data(bubbleSteps)
      .join(
        enter => enter
          .append('path')
          .attr('fill', 'transparent')
          .attr('stroke-width', d => onePixel * maxVal / max)
          .attr('stroke', '#ccc')
          .attr('transform', d => `translate(${max}, -${d}) scale(${maxVal ? (1 * d / maxVal) : 1})`)
          .attr('d', form),
        update => update
          .attr('fill', 'transparent')
          .attr('stroke-width', d => onePixel * maxVal / max)
          .attr('stroke', '#ccc')
          .attr('transform', d => `translate(${max}, -${d}) scale(${maxVal ? (1 * d / maxVal) : 1})`)
          .attr('d', form),
        exit => exit
          .remove()
      )

      appendOnce(legend, 'g', 'legendSelectedBubble')
        .selectAll('path')
        .data(data.filter(o => selectedId && defs.bubble && defs.itemId(o) === selectedId), defs.itemId)
        .join(
          enter => enter
            .append('path')
            .attr('fill', 'transparent')
            .attr('stroke-width', d => maxVal ? (onePixel * maxVal / max * 2) : 0)
            .attr('stroke', '#000')
            .attr('transform', d => `translate(${max}, -${bubbleScala(defs.bubble(d))}) scale(${maxVal ? (1 * bubbleScala(defs.bubble(d)) / maxVal) : 1})`)
            .attr('d', form),
          update => update
            .attr('fill', 'transparent')
            .attr('stroke-width', d => maxVal ? (onePixel * maxVal / max * 2) : 0)
            .attr('stroke', '#000')
            .attr('transform', d => `translate(${max}, -${bubbleScala(defs.bubble(d))}) scale(${maxVal ? (1 * bubbleScala(defs.bubble(d)) / maxVal) : 1})`)
            .attr('d', form),
          exit => exit
            .remove()
        )

      if (defs.bubbleLabel !== undefined) {
        const bg = appendOnce(legend, 'rect', 'legendSelectedBubbleLabelBg')
          .attr('x', max * 2)
          .attr('y', -fontSize-5)
          .attr('fill', '#ffffffdd')
          .attr('height', fontSize + 10)

        const label = appendOnce(legend, 'text', 'legendSelectedBubbleLabel')
          .attr('font-size', fontSize)
          .attr('x', max * 2 + 5)
          .attr('y', -5)
          .text(data
            .filter(o => selectedId && defs.bubble && defs.bubbleLabel && defs.itemId(o) === selectedId)
            .map(o => defs.bubbleLabel ? defs.bubbleLabel(o) : 0)
          )

        bg.attr('width', label.node().getBBox().width + 10)
      }

  }

  // bg
  appendOnce(legend, 'rect', 'bg')
    .attr('fill', '#ffffffdd')
    .attr('width', outerWidth)
    .attr('height', outerHeight)

  if (defs.legendGroup) {
    const space = onePixel * 6
    const box = onePixel * 10
    for (const range of defs.ranges) {
      let i = 0
      let y = padding
      let left = padding
      for (const color of range.colors) {
        const label = (range.ticks || range.labels || range.values)[i]
        // rect
        appendOnce(legend, 'rect', `dot-${i}`)
          .attr('fill', color)
          .attr('x', left)
          .attr('y', y + (fontSize - box))
          .attr('width', box)
          .attr('height', box)
        left += box + space
        // label
        const labelNode = appendOnce(legend, 'text', `label-${i}`)
          .attr('font-weight', 'bold')
          .attr('dominant-baseline', 'baseline')
          .attr('fill', '#000')
          .attr('font-size', fontSize)
          .text(label)
          .attr('x', left)
          .attr('y', y + fontSize)
        left += labelNode.node().getBBox().width + space * 2
        if (left > innerWidth) {
          left = padding
          y += fontSize + padding
        }
        i++
      }
    }
  } else {
    let i = 0
    for (const range of defs.ranges) {
      addGradient(d3, defsNode, range.colors, id, i)
      i++
    }

    const hasNullValue = data.map(defs.value).find(o => o === null) !== undefined
    const hasStripedValue = defs.striped && data.map(defs.striped).find(o => o) !== undefined

    // null
    let specialBarsWidth = 0
    if (hasNullValue) {
      const nullBar = appendOnce(legend, 'rect', 'null')
        .attr('fill', '#000')
        .attr('y', barTop)
        .attr('height', barHeight).attr('stroke', '#fff')
        .attr('stroke-width', onePixel * 2)
      const nullLabel = appendOnce(legend, 'text', `null-label`)
        .attr('fill', '#000')
        .attr('font-size', fontSize)
        .attr('font-weight', 'normal')
        .attr('dominant-baseline', !hasStripedValue ? 'auto' : 'hanging')
        .text(() => onePixel > 1 ? 'k.A.' : 'keine Angaben')
        .attr('y', !hasStripedValue ? textBottom : tickTop)
      const w = nullLabel.node().getBBox().width + 4
      specialBarsWidth += w
      nullBar
        .attr('width', w)
        .attr('x', innerWidth - specialBarsWidth)
      nullLabel
        .attr('x', innerWidth - specialBarsWidth + 2)
    }

    // stripes
    if (hasStripedValue) {
      addGradient(d3, defsNode, defs.ranges.map(o => o.colors).flat(), id, 'stripe')
      const stripedBar = appendOnce(legend, 'rect', 'stripes')
        .attr('fill', `url(#gradient-${id}-stripe)`)
        .attr('mask', `url(#mask-stripe-${id})`)
        .attr('y', barTop)
        .attr('height', barHeight)
        .attr('stroke', '#fff')
        .attr('stroke-width', onePixel * 2)
      const stripedLabel = appendOnce(legend, 'text', `stripes-label`)
        .attr('fill', '#000')
        .attr('font-size', fontSize)
        .attr('font-weight', 'normal')
        .text(() => onePixel > 1 ? 'n.a.' : 'nicht abgeschlossen')
        .attr('y', textBottom)
      const w = stripedLabel.node().getBBox().width + 4
      specialBarsWidth += w
      stripedBar
        .attr('width', w)
        .attr('x', innerWidth - specialBarsWidth)
      stripedLabel
        .attr('x', innerWidth - specialBarsWidth + 2)
    }

    const rangesWidth = innerWidth - specialBarsWidth
    const barWidth = rangesWidth / defs.ranges.length

    i = 0
    for (const range of defs.ranges) {
      const barLeft = i * barWidth + left
      // bar
      appendOnce(legend, 'rect', `gradient-${id}-${i}`)
        .attr('id', `bar-${id}-${i}`)
        .attr('fill', `url(#gradient-${id}-${i})`)
        .attr('x', barLeft)
        .attr('y', barTop)
        .attr('width', barWidth)
        .attr('height', barHeight)
        .attr('stroke', '#fff')
        .attr('stroke-width', onePixel * 2)

      if (range.label) {
        // label
        appendOnce(legend, 'text', `label-${i}`)
          .attr('text-anchor', i === 0 ? 'start' : (i === defs.ranges.length - 1 ? 'end' : 'middle'))
          .attr('font-family', 'HelveticaNowBlack, Helvetica, sans-serif')
          .attr('fill', range.labelColor || darkestHexColor(range.colors))
          .attr('font-size', fontSize)
          .text(range.label)
          .attr('y', textBottom)
          .attr('x', i === 0 ? barLeft : (i === defs.ranges.length - 1 ? barLeft + barWidth - 2 : barLeft + barWidth / 2))
      }

      function values () {
        return range.ticks || range.values
      }

      // ticks
      function shiftLabelX (d, index) {
        const tickValues = values()
        // prev ticks
        if (i > 0) {
          const prevRange = defs.ranges[i - 1]
          const prevTickValues = (prevRange.ticks || prevRange.values)
          // if last tick from prev range the same value as first tick of actual range => center it
          if (index === 0 && prevTickValues && prevTickValues[prevTickValues.length - 1] === tickValues[0]) {
            return barLeft - this.getBBox().width * 0.5
          }
        }
        const shift = 1 / (tickValues.length - 1) * index
        return barLeft + margin + (barWidth - 2 * margin) * shift - this.getBBox().width * shift
      }
      function labelOpacity (d, index) {
        const tickValues = values()
        // next ticks
        if (i < defs.ranges.length - 1) {
          const nextRange = defs.ranges[i + 1]
          const nextTickValues = (nextRange.ticks || nextRange.values)
          // if first tick from next range the same value as last tick of actual range => hide it
          if (index === tickValues.length - 1 && nextTickValues[0] === tickValues[tickValues.length - 1]) {
            return 0
          }
        }
        return 1
      }

      const ticks = legend.selectAll(`.tick-${i}`).data(values)
      ticks.join(
        enter => enter
          .append('text')
          .attr('class', `tick-${i}`)
          .attr('font-weight', 'normal')
          .attr('dominant-baseline', 'hanging')
          .attr('fill', '#000')
          .attr('opacity', labelOpacity)
          .attr('font-size', fontSize)
          .text(d => typeof d === 'string' ? d : d.toFixed(1).replace(/.0$/, ''))
          .attr('x', shiftLabelX)
          .attr('y', tickTop),
        update => update
          .attr('font-size', fontSize)
          .attr('opacity', labelOpacity)
          .text(d => typeof d === 'string' ? d : d.toFixed(1).replace(/.0$/, ''))
          .attr('x', shiftLabelX)
          .attr('y', tickTop),
      )

      i++
    }
  }

  // selector active
  appendOnce(legend, 'circle', `sel-active`)
    .attr('id', `sel-active-${id}`)
    .attr('fill', '#000')
    .attr('cy', barTop + onePixel * 6)
    .attr('r', onePixel * 6)
    .attr('stroke', '#fff')
    .attr('stroke-width', onePixel * 2)
    .attr('opacity', 0)
  // selector active label
  appendOnce(legend, 'text', `sel-active-label`)
    .attr('id', `sel-active-${id}-label`)
    .attr('fill', '#000')
    .attr('dominant-baseline', 'hanging')
    .attr('text-anchor', 'middle')
    .attr('font-size', fontSize)
    .attr('y', barTop + onePixel * 18)
    .attr('opacity', 0)

  // selector
  appendOnce(legend, 'circle', `selector`)
    .attr('id', `selector-${id}`)
    .attr('fill', '#fff')
    .attr('cy', barTop + onePixel * 6)
    .attr('r', onePixel * 6)
    .attr('stroke', '#000')
    .attr('stroke-width', onePixel * 2)
    .attr('opacity', 0)

  return top
}

const addTooltip = (root) => {
  const tooltip = root.append('div')
  root.attr('style', 'position:relative;')
  tooltip.attr('style', 'opacity:0; z-index:9999; background-color:white; padding:16px; pointer-events:none; position:absolute; box-shadow:0 0 16px #00000066;')
  return tooltip
}

const download = (reset, staticNode, svg, title, width, height) => {
  // reset zoom
  reset(0)

  setTimeout(() => {
    const download = staticNode.select('.download')
    const plus = staticNode.select('.plus')
    const minus = staticNode.select('.minus')
    // hide buttons
    download.attr('opacity', 0)
    plus.attr('opacity', 0)
    minus.attr('opacity', 0)
    
    let filename = document.location.pathname.substring(1).replace(/\//g, '-')
    if (title) {
      filename += '-' + title.replace(/ /g, '_')
    }
    // title
    const titleNode = appendOnce(svg, 'text', 'title')
      .attr('fill', '#000')
      .attr('alignment-baseline', 'baseline')
      .text(title)
      .attr('font-size', '20px')
      .attr('x', 10)
      .attr('y', 0)

    svg.attr('viewBox', [0, -30, width, height + 30])
    downloadDataURL(svg.node(), filename)
    svg.attr('viewBox', [0, 0, width, height])
    titleNode.remove()

    // show buttons again
    download.attr('opacity', 1)
    plus.attr('opacity', 1)
    minus.attr('opacity', 1)
  }, 1)
}

const svgDataURL = (svg) => {
  const svgAsXML = (new XMLSerializer()).serializeToString(svg)
  return 'data:image/svg+xml,' + encodeURIComponent(svgAsXML)
}

const addStripePattern = (defs, id) => {
  const size = 16
  const pattern = appendOnce(defs, 'pattern', 'pattern-stripe')
    .attr('id', `pattern-stripe-${id}`)
    .attr('width', size)
    .attr('height', size)
    .attr('patternUnits', 'userSpaceOnUse')
    .attr('patternTransform', 'rotate(-30)')
  appendOnce(pattern, 'rect', 'one')
    .attr('width', size / 2)
    .attr('height', size)
    .attr('fill', '#bbb')
  appendOnce(pattern, 'rect', 'two')
    .attr('x', size / 2)
    .attr('width', size / 2)
    .attr('height', size)
    .attr('fill', '#fff')
  const mask = appendOnce(defs, 'mask', 'mask')
    .attr('id', `mask-stripe-${id}`)
  appendOnce(mask, 'rect', 'bounds')
    .attr('x', '0')
    .attr('y', '0')
    .attr('width', '100%')
    .attr('height', '100%')
    .attr('fill', `url(#pattern-stripe-${id})`)
}

const downloadDataURL = (svg, name) => {
  const dl = document.createElement('a')
  document.body.appendChild(dl) // This line makes it work in Firefox.
  dl.setAttribute('href', svgDataURL(svg))
  dl.setAttribute('download', `${name}.svg`)
  dl.click()
}

const addDownload = (d3, node, callback, size = 20) => {
  const padding = 6
  // download
  const download = appendOnce(node, 'g', 'download')
  download.attr('transform', 'translate(-100, -100)')
  download.append('rect')
    .attr('width', size)
    .attr('height', size)
    .attr('fill', '#ccc')
    .on('mouseover', function (event, d) {
      const item = d3.select(this)
      item.attr('style', 'filter:contrast(50%) brightness(140%); cursor:pointer;')
    })
    .on('mouseout', function (event, d) {
      const item = d3.select(this)
      item.attr('style', '')
    })
    .on('click', function (event) {
      event.stopPropagation()
      callback()
    })

  download.append('path')
    .attr('pointer-events', 'none')
    .attr('transform', `scale(0.45 0.45) translate(${padding * 1.6} ${padding * 1.6})`)
    .attr('d', 'M18.4459 11.6445L16.8041 10.1055L13.125 14.0302V0.75H10.875V13.8997L7.21838 9.756L5.53163 11.2444L11.9749 18.5467L18.4459 11.6445Z M2 21H22V23H2V21Z')
    .attr('fill', '#000')
}

const addZoomIn = (d3, node, callback, size = 20) => {
  const padding = 6
  // plus
  const plus = appendOnce(node, 'g', 'plus')
  plus.attr('transform', 'translate(-100, -100)')
  plus.append('rect')
    .attr('width', size)
    .attr('height', size)
    .attr('fill', '#ccc')
    .attr('cursor', 'pointer')
    .on('mouseover', function (event, d) {
      const item = d3.select(this)
      item.attr('style', 'filter:contrast(50%) brightness(140%);')
    })
    .on('mouseout', function (event, d) {
      const item = d3.select(this)
      item.attr('style', '')
    })
    .on('click', function (event) {
      event.stopPropagation()
      callback()
    })

  plus.append('rect')
    .attr('pointer-events', 'none')
    .attr('x', padding)
    .attr('y', (size - 1) / 2)
    .attr('width', size - 2 * padding)
    .attr('height', 1)
    .attr('fill', '#000')

  plus.append('rect')
    .attr('pointer-events', 'none')
    .attr('x', (size - 1) / 2)
    .attr('y', padding)
    .attr('width', 1)
    .attr('height', size - 2 * padding)
    .attr('fill', '#000')
  return plus
}

const addZoomOut = (d3, node, callback, size = 20) => {
  const padding = 6
  // minus
  const minus = appendOnce(node, 'g', 'minus')
  minus.attr('transform', 'translate(-100, -100)')
  minus.append('rect')
    .attr('width', size)
    .attr('height', size)
    .attr('fill', '#666')
    .on('mouseover', function (event, d) {
      const item = d3.select(this)
      item.attr('style', 'filter:contrast(50%) brightness(140%);')
    })
    .on('mouseout', function (event, d) {
      const item = d3.select(this)
      item.attr('style', '')
    })
    .on('click', function (event) {
      event.stopPropagation()
      callback()
    })

  minus.append('rect')
    .attr('pointer-events', 'none')
    .attr('x', padding)
    .attr('y', (size - 1) / 2)
    .attr('width', size - 2 * padding)
    .attr('height', 1)
    .attr('fill', '#000')
  return minus
}

const map = async (d3, target, data = [], defs = {}, selectedId = null, width = 600, height = 600) => {
  const id = d3.selectAll('.svg').nodes().length
  let selectedCallback = null
  let hoverId = null
  const root = d3.select(target)

  const svg = appendOnce(root, 'svg', 'svg')
    .attr('viewBox', [0, 0, width, height])
    .attr('font-family', 'HelveticaNowRegular, Helvetica, sans-serif')

  const zoomMargin = 0.05
  let zoom = null
  let zoomFactor = 1
  let plus = null
  let minus = null
  const zoomNode = appendOnce(svg, 'g', 'zoom-area')
  const staticNode = appendOnce(svg, 'g', 'static-area')

  const zoomInFunc = () => {
    zoomFactor *= 2
    zoomFactor = Math.min(zoomFactor, 4)
    svg.transition().remove()
    setTimeout(() => {
      svg.transition().call(zoom.scaleTo, zoomFactor - zoomMargin)
    }, 0)
  }
  const zooOutFunc = () => {
    zoomFactor /= 2
    zoomFactor = Math.max(zoomFactor, 1)
    svg.transition().remove()
    setTimeout(() => {
      svg.transition().call(zoom.scaleTo, zoomFactor - zoomMargin)
    }, 0)
  }

  if (defs.zoomable) {
    plus = addZoomIn(d3, staticNode, zoomInFunc)
    minus = addZoomOut(d3, staticNode, zooOutFunc)
  }

  if (defs.download) {
    addDownload(d3, staticNode, () => {
      download(reset, staticNode, svg, defs.downloadTitle, width, height)
    })
  }

  const bounds = {
    type: 'FeatureCollection',
    features: []
  }

  if (!defs.url) {
    console.error('url prop is requied data item!')
  }
  const geojson = await d3.json(defs.url)
  const itemIds = data.map(o => defs.itemId(o))
  bounds.features.push(...geojson.features.filter(o => {
    const actualId = defs.mapId(o.properties)
    if (defs.showOnUndefined) {
      return defs.showOnUndefined(actualId)
    }
    return itemIds.includes(actualId)
  }))

  let bg = null
  if (defs.bg) {
    bg = await d3.json(defs.bg)
    bounds.features.push(...bg.features)
  }

  const defsNode = appendOnce(svg, 'defs', 'defs')
  if (defs.striped) {
    addStripePattern(defsNode, id)
  }

  let top = height
  if (defs.legend) {
    top = height - 70 * getOnePixel(svg, width)
  }

  const projection = d3.geoIdentity().reflectY(true).fitSize([width, top], bounds)
  const path = d3.geoPath().projection(projection)
  const geometry = d => {
    const found = bounds.features.find(o => defs.mapId(o.properties) === defs.itemId(d))
    if (!found) {
      return null
    }
    return path(found.geometry)
  }
  const rect = d => {
    const found = bounds.features.find(o => defs.mapId(o.properties) === defs.itemId(d))
    if (!found) {
      return { left: -Infinity, top: -Infinity, right: -Infinity, bottom: -Infinity }
    }
    const rect = path.bounds(found.geometry)
    if (!rect) {
      return { left: -Infinity, top: -Infinity, right: -Infinity, bottom: -Infinity }
    }
    return { left: rect[0][0], top: rect[0][1], right: rect[1][0], bottom: rect[1][1] }
  }
  const domain = defs.ranges.map(o => o.values).flat()
  const range = defs.ranges.map(o => o.colors).flat()
  const color = d3
    .scaleLinear()
    .domain(domain.length === 1 ? [domain[0], domain[0]] : domain)
    .range(range.length === 1 ? [range[0], range[0]] : range)
    .unknown('#000')

  const bgNode = appendOnce(zoomNode, 'g', 'bg')
  if (defs.bg) {
    bgNode.selectAll('path.bg')
      .data(bg.features)
      .enter()
      .append('path')
      .attr('class', 'bg')
      .attr('d', path)
      .attr('fill', '#cee6f4')
      .attr('stroke', '#fff')
      .attr('stroke-width', 2)
  }

  if (defs.bubble !== undefined || defs.showOnUndefined) {
    bgNode.selectAll('path.forms')
      .data(geojson.features.filter(o => {
        const actualId = defs.mapId(o.properties)
        if (defs.showOnUndefined) {
          return defs.showOnUndefined(actualId)
        }
        return itemIds.includes(actualId)
      }))
      .enter()
      .append('path')
      .attr('class', 'forms')
      .attr('d', path)
      .attr('fill', '#eee')
      .attr('stroke', '#fff')
      .attr('stroke-width', 2)
  }

  let tooltip = null
  if (defs.tooltip) {
    tooltip = addTooltip(root)
  }

  const updateSelector = (key, value) => {
    const selector = svg.select(key)
    if (!selector.empty()) {

      const range = defs.ranges.find(o => o.values[0] <= value && value <= o.values[o.values.length - 1])
      const index = defs.ranges.indexOf(range)

      const bar = svg.select(`#bar-${id}-${index}`)
      if (!bar.empty()) {
        const widths = [...Array(range.values.length)]
        .map((_, i) => 1 / (range.values.length - 1) * i) // interpolate 0 - 1
        .map(o => o * parseFloat(bar.attr('width')))
        const x = d3
          .scaleLinear()
          .domain(range.values)
          .range(widths)
        const left = parseFloat(bar.attr('x')) + x(value)
        // dot
        selector.transition(100)
          .attr('cx', left)
          .attr('opacity', 1)
        // label
        svg.select(key + '-label')
          .attr('x', left)
          .text(value)
          .attr('opacity', 1)
      }
    }
  }
  const hideSelector = (key) => {
    const selector = svg.select(key)
    if (!selector.empty()) {
      // dot
      selector.transition(100)
        .attr('opacity', 0)
      // label
      svg.select(key + '-label')
        .attr('opacity', 0)
    }
  }

  const mouseOver = function (event, d) {
    const newId = defs.itemId(d)
    hoverId = newId
    const item = d3.select(this)
    item.attr('stroke', '#000')
    item.attr('fill', 'rgba(255,255,255,0.3)')

    const value = defs.value(d)
    if (value !== null) {
      updateSelector(`#selector-${id}`, value)
    }

    // item.raise()
    if (defs.tooltip) {
      tooltip
        .html(defs.tooltip(d))
        .style('opacity', 1)
      mouseMove(event, d)
    }
  }

  const mouseMove = function (event, d) {
    if (defs.tooltip) {
      const pos = root.node().getBoundingClientRect()
      const tt = tooltip.node().getBoundingClientRect()
      const center = pos.x + pos.width / 2
      let x = event.clientX - pos.x + 20
      if (event.clientX > center) {
        x = event.clientX - pos.x - 20 - tt.width
      }
      x = Math.max(x, 0)
      x = Math.min(x, document.body.clientWidth - tt.width - 20)

      const y = Math.max(0, event.clientY - pos.y - tt.height - 20)
      tooltip
        .style('left', x + 'px')
        .style('top', y + 'px')
    }
  }

  const mouseOut = function (event, d) {
    hoverId = null
    const item = d3.select(this)
    item.attr('stroke', 'transparent')
    item.attr('fill', null)

    hideSelector(`#selector-${id}`)

    if (defs.tooltip) {
      tooltip
        .style('opacity', 0)
    }
  }

  const zoomToItem = nr => {
    const item = d3.select(`#item-${id}-${nr}`)
    const bBox = item.node().getBBox()
    svg.transition().duration(750).call(
      zoom.transform,
      d3.zoomIdentity
        .translate(width / 2, height / 2)
        .scale(0.6 * Math.min(width / bBox.width, height / bBox.height))
        .translate(-bBox.x - bBox.width / 2, -bBox.y - bBox.height / 2),
    )
  }

  const click = function (event, d) {
    event.stopPropagation()
    const newId = defs.itemId(d)

    if (defs.selectable) {
      if (selectedId !== newId) {
        if (selectedCallback) {
          selectedCallback(selectedId === newId ? null : newId)
        }
      } else {
        if (defs.unselectable) {
          if (selectedCallback) {
            selectedCallback(null)
          }
        }
      }
    }
  }

  const colorBgNode = appendOnce(zoomNode, 'g', 'color-bg')
    .attr('stroke', '#fff')
    .attr('stroke-width', 1)
    .attr('pointer-events', 'none')
    .attr('text-anchor', 'middle')
    .attr('font-size', '14px')

  const groupNode = appendOnce(zoomNode, 'g', 'group-bg')
    .attr('stroke', '#fff')
    .attr('fill', 'transparent')
    .attr('stroke-width', 4)
    .attr('pointer-events', 'none')

  const colorNode = appendOnce(zoomNode, 'g', 'color')
    .attr('stroke', '#fff')
    .attr('stroke-width', 1)
    .attr('pointer-events', 'none')
    .attr('text-anchor', 'middle')
    .attr('font-size', '12px')

  let hoverLineNode = null
  let selectionNode = null
  if (defs.hoverable) {
    if (defs.selectable) {
      selectionNode = appendOnce(zoomNode, 'g', 'selection')
        .attr('pointer-events', 'none')
    }
    hoverLineNode = appendOnce(zoomNode, 'g', 'hover-line')
      .attr('stroke', 'transparent')
      .attr('fill', 'transparent')
      .attr('stroke-width', 1)
  }

  const update = () => {
    if (defs.legend || defs.zoomable) {
      screenSize(d3, svg, staticNode, defsNode, data, defs, width, height, id, selectedId)
    }

    // path or bubble
    colorBgNode.selectAll('path')
      .data([...data]
        .filter(o => !defs.bubble && defs.groupIds ? !defs.groupIds.includes(defs.itemId(o)) : false)
        .sort((a, b) => defs.bubble ? d3.ascending(defs.bubble(b) || Infinity, defs.bubble(a) || Infinity) : 0),
        defs.itemId)
      .join(
        enter => enter
          .append('path')
          .attr('id', d => `item-${id}-${defs.itemId(d)}`)
          .attr('d', d => defs.bubble && defs.value(d) !== null ? bubbleGeometry(defs, center(bounds, path, defs, d), d) : geometry(d))
          .attr('fill', d => defs.color ? defs.color(d, color(defs.value(d))) : color(defs.value(d)))
          .attr('mask', d => defs.striped && defs.striped(d) ? `url(#mask-stripe-${id})` : ''),
        update => update
          .attr('id', d => `item-${id}-${defs.itemId(d)}`)
          .attr('d', d => defs.bubble && defs.value(d) !== null ? bubbleGeometry(defs, center(bounds, path, defs, d), d) : geometry(d))
          .attr('fill', d => defs.color ? defs.color(d, color(defs.value(d))) : color(defs.value(d)))
          .attr('mask', d => defs.striped && defs.striped(d) ? `url(#mask-stripe-${id})` : ''),
        exit => exit
          .remove()
      )

    // bold outline bg
    groupNode.selectAll('path')
      .data([...data]
        .filter(o => !defs.bubble && defs.groupIds ? defs.groupIds.includes(defs.itemId(o)) : false),
        defs.itemId)
      .join(
        enter => enter
          .append('path')
          .attr('d', d => defs.bubble && defs.value(d) !== null ? bubbleGeometry(defs, center(bounds, path, defs, d), d) : geometry(d)),
        update => update
          .attr('d', d => defs.bubble && defs.value(d) !== null ? bubbleGeometry(defs, center(bounds, path, defs, d), d) : geometry(d)),
        exit => exit
          .remove()
      )

    // path or bubble
    colorNode.selectAll('path')
      .data([...data]
        .filter(o => !defs.bubble && defs.groupIds ? defs.groupIds.includes(defs.itemId(o)) : true)
        .sort((a, b) => defs.bubble ? d3.ascending(defs.bubble(b) || Infinity, defs.bubble(a) || Infinity) : 0),
        defs.itemId)
      .join(
        enter => enter
          .append('path')
          .attr('id', d => `item-${id}-${defs.itemId(d)}`)
          .attr('d', d => defs.bubble && defs.value(d) !== null ? bubbleGeometry(defs, center(bounds, path, defs, d), d) : geometry(d))
          .attr('fill', d => defs.color ? defs.color(d, color(defs.value(d))) : color(defs.value(d)))
          .attr('mask', d => defs.striped && defs.striped(d) ? `url(#mask-stripe-${id})` : ''),
        update => update
          .attr('id', d => `item-${id}-${defs.itemId(d)}`)
          .attr('d', d => defs.bubble && defs.value(d) !== null ? bubbleGeometry(defs, center(bounds, path, defs, d), d) : geometry(d))
          .attr('fill', d => defs.color ? defs.color(d, color(defs.value(d))) : color(defs.value(d)))
          .attr('mask', d => defs.striped && defs.striped(d) ? `url(#mask-stripe-${id})` : ''),
        exit => exit
          .remove()
      )

    // label
    if (defs.label) {
      colorNode.selectAll('text')
        .data([...data].filter(o => defs.label(o)), defs.itemId)
        .join(
          enter => enter
            .append('text')
            .attr('fill', '#000')
            .attr('stroke-width', 0)
            .attr('x', d => center(bounds, path, defs, d).x)
            .attr('y', d => rect(d).top - 5)
            .text(d => defs.label(d)),
          update => update
            .attr('x', d => center(bounds, path, defs, d).x)
            .attr('y', d => rect(d).top - 5)
            .text(d => defs.label(d)),
          exit => exit
            .remove()
        )
    }

    // hover
    if (defs.hoverable) {
      const hoverNodes = hoverLineNode.selectAll(`path`)
        .data(data, defs.mapId)
        .join(
          enter => enter
            .append('path')
            .attr('d', geometry),
          update => update
            .attr('d', geometry),
          exit => exit
            .remove()
        )

      hoverNodes
        .on('mouseover', mouseOver)
        .on('mousemove', mouseMove)
        .on('mouseout', mouseOut)

      // select
      if (defs.selectable) {
        hoverNodes
          .on('click', click)
          .style('cursor', 'pointer')
      }
    }

    function getBBox (d) {
      d.bbox = this.getBBox()
    }

    if (defs.selectable) {
      // wait for item is rendered
      const steps = 6
      const border = 4
      const size = 2
      // path
      selectionNode.selectAll('g')
        .data(data.filter(o => defs.itemId(o) === selectedId), defs.itemId)
        .join(
          enter => {
            const group = enter.append('g')

            // shadow
            for (let i = 1; i < steps; i++) {
              group
                .append('path')
                .attr('d', d => defs.bubble ? bubbleGeometry(defs, center(bounds, path, defs, d), d) : geometry(d))
                .attr('stroke-width', i * size + border)
                .attr('stroke', '#00000022')
            }
            // white outline
            group
              .append('path')
              .attr('d', d => defs.bubble ? bubbleGeometry(defs, center(bounds, path, defs, d), d) : geometry(d))
              .attr('stroke-width', border)
              .attr('stroke', '#fff')
            // form
            group
              .append('path')
              .attr('d', d => defs.bubble ? bubbleGeometry(defs, center(bounds, path, defs, d), d) : geometry(d))
              .attr('fill', d => defs.color ? defs.color(d, color(defs.value(d))) : color(defs.value(d)))
              .attr('mask', d => defs.striped && defs.striped(d) ? `url(#mask-stripe-${id})` : '')
          },
          exit => exit
            .remove()
        )
      // label
      if (defs.selectedLabel) {
        selectionNode.selectAll('text')
          .data(data.filter(o => defs.itemId(o) === selectedId && (!defs.label || !defs.label(o))), defs.itemId)
          .join(
            enter => enter
                .append('text')
                .attr('text-anchor', 'middle')
                .attr('dominant-baseline', 'middle')
                .attr('font-family', 'HelveticaNowBlack, Helvetica, sans-serif')
                .attr('fill', '#fff')
                .attr('paint-order', 'stroke')
                .attr('stroke-width', '2px')
                .attr('stroke', '#000')
                .attr('stroke-linecap', 'butt')
                .attr('stroke-linejoin', 'miter')
                .text(defs.selectedLabel)
                .each(getBBox)
                .attr('x', d => holdInBounds(bounds, path, center(bounds, path, defs, d), d).x)
                .attr('y', d => holdInBounds(bounds, path, center(bounds, path, defs, d), d).y),
            exit => exit
              .remove()
          )
        }
    }
  }
  update()

  function zoomed ({ transform }) {
    // toggle zoom on / off
    if(transform.k === 1 - zoomMargin) {
      svg.on('.zoom', null)
    } else {
      svg.call(zoom)
    }

    // toggle bg color
    if (plus) {
      plus.select('rect:first-child')
        .attr('fill', transform.k === 4 - zoomMargin ? '#666' : '#ccc')
        .attr('cursor', transform.k === 4 - zoomMargin ? 'auto' : 'pointer')
    }
    if (minus) {
      minus.select('rect:first-child')
        .attr('fill', transform.k === 1 - zoomMargin ? '#666' : '#ccc')
        .attr('cursor', transform.k === 1 - zoomMargin ? 'auto' : 'pointer')
    }

    zoomNode.attr('transform', transform)
  }

  zoom = d3.zoom()
    .scaleExtent([1 - zoomMargin, 4 - zoomMargin])
    .translateExtent([[0, 0], [width, height]]) // limit drag
    .on('zoom', zoomed)

  function reset (duration = 750) {
    svg.transition().duration(duration).call(
      zoom.transform,
      d3.zoomIdentity,
      d3.zoomTransform(svg.node()).invert([width / 2, height / 2])
    )
  }

  svg.call(zoom.scaleTo, zoomFactor - 1 - zoomMargin)

  return {
    selectedId: (selId) => {
      if (selId && selectedId !== selId) {
        const item = selId !== null ? data.find(o => defs.itemId(o) === selId) : null
        if (!item) {
          console.error(`Id: ${selId} not found in data !`)
          return
        }
        selectedId = selId
        updateSelector(`#sel-active-${id}`, defs.value(item))
      } else {
        selectedId = null
        hideSelector(`#sel-active-${id}`)
        // reset()
      }
      update()
    },
    selectedIdCallback: (callbackFunction) => {
      selectedCallback = callbackFunction
    },
    updateData: (_data) => {
      data = _data
      update()
    },
    updateDefs: (_defs) => {
      defs = _defs
      update()
    },
    update,
    reset,
    zoomIn: () => {
      if (defs.zoomable) {
        zoomInFunc()
      }
    },
    zoomOut: () => {
      if (defs.zoomable) {
        zooOutFunc()
      }
    },
    download: (title) => {
      download(reset, staticNode, svg, title, width, height)
    }
  }
}

export {
  map
}
